# -*- coding: utf-8 -*-
"""
Insert document views
"""
import json

from django.core.urlresolvers import reverse
from django.http import HttpResponse
from django.views import generic

from braces.views import PermissionRequiredMixin

from djangocodemirror.views import SampleQuicksaveMixin

from rstview.parser import SourceParser

from sveedocuments.models import Insert
from sveedocuments.forms.insert import InsertForm, InsertEditForm, InsertQuickForm
from sveedocuments.utils.objects import get_instance_children

class InsertCreateView(PermissionRequiredMixin, generic.CreateView):
    """
    Form view to create an *Insert* document
    """
    model = Insert
    context_object_name = "insert_instance"
    template_name = "sveedocuments/board/insert_form.html"
    form_class = InsertForm
    permission_required = "sveedocuments.add_insert"
    raise_exception = True
    _redirect_to_self = False

    def post(self, request, *args, **kwargs):
        # Mark to go back to the form page after save, triggered by special submit
        if request.POST:
            if request.POST.get('submit_and_continue', False):
                self._redirect_to_self = True
        return super(InsertCreateView, self).post(request, *args, **kwargs)

    def get_success_url(self):
        if self._redirect_to_self:
            return reverse('sveedocuments:insert-edit', args=[self.object.slug])
        return reverse('sveedocuments:insert-index')
    
    def get_form_kwargs(self):
        kwargs = super(InsertCreateView, self).get_form_kwargs()
        kwargs.update({
            'author': self.request.user,
        })
        return kwargs

class InsertEditView(PermissionRequiredMixin, generic.UpdateView):
    """
    Form view to edit an *Insert* document
    """
    model = Insert
    context_object_name = "insert_instance"
    template_name = "sveedocuments/board/insert_form.html"
    form_class = InsertEditForm
    permission_required = "sveedocuments.change_insert"
    raise_exception = True
    _redirect_to_self = False

    def post(self, request, *args, **kwargs):
        # Mark to go back to the form page after save, triggered by special submit
        if request.POST:
            if request.POST.get('submit_and_continue', False):
                self._redirect_to_self = True
        return super(InsertEditView, self).post(request, *args, **kwargs)

    def get_success_url(self):
        if self._redirect_to_self:
            return reverse('sveedocuments:insert-edit', args=[self.object.slug])
        return reverse('sveedocuments:insert-index')
    
    def get_form_kwargs(self):
        kwargs = super(InsertEditView, self).get_form_kwargs()
        kwargs.update({'author': self.request.user})
        return kwargs

class InsertDeleteView(PermissionRequiredMixin, generic.DeleteView):
    """
    Form view to delete an *Insert* document
    
    This display a tree of relations to confirm the object and relations deletion
    """
    model = Insert
    context_object_name = "insert_instance"
    template_name = "sveedocuments/board/insert_delete.html"
    permission_required = "sveedocuments.delete_insert"
    raise_exception = True

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        context = self.get_context_data(object=self.object)
        context.update({'relations': get_instance_children(self.object)})
        return self.render_to_response(context)

    def get_success_url(self):
        return reverse('sveedocuments:insert-index')

class InsertQuicksaveView(SampleQuicksaveMixin, InsertEditView):
    """
    Quicksave view for an *Insert* content
    """
    form_class = InsertQuickForm
    
    def get_object(self, queryset=None):
        if self.request.POST.get('slug', False):
            self.kwargs['slug'] = self.request.POST['slug']
        return super(InsertQuicksaveView, self).get_object(queryset=queryset)



# -*- coding: utf-8 -*-
"""
Root url's map for application
"""
from django.conf.urls import *

from sveedocuments.views.page import (HelpPageView, PageIndexView, PageDetailsView, 
                                        PageSourceView, PagePDFView)

urlpatterns = patterns('',
    url(r'^$', PageIndexView.as_view(), name='index'),
    
    (r'^board/', include('sveedocuments.urls_board')),
    
    url(r'^sitemap/$', PageIndexView.as_view(), name='index'),
    url(r'^help/$', HelpPageView.as_view(), name='help'),
    
    url(r'^(?P<slug>[-\w]+)/$', PageDetailsView.as_view(), name='page-details'),
    url(r'^(?P<slug>[-\w]+)/source/$', PageSourceView.as_view(), name='page-source'),
)

if not getattr(PagePDFView, 'is_dummy', False):
    urlpatterns += patterns('',
        url(r'^(?P<slug>[-\w]+)/pdf/$', PagePDFView.as_view(), name='page-pdf'),
    )



# -*- coding: utf-8 -*-
"""
Insert forms
"""
from django import forms
from django.utils.translation import ugettext as _

from djangocodemirror.fields import CodeMirrorWidget

from rstview.parser import SourceReporter, map_parsing_errors

from sveedocuments.local_settings import DOCUMENTS_PAGE_RESERVED_SLUGS
from sveedocuments.models import Insert, documents_insert_update_signal
from sveedocuments.forms import CrispyFormMixin
from sveedocuments.forms.page import PageQuickForm

class InsertForm(CrispyFormMixin, forms.ModelForm):
    """
    Insert form
    """
    crispy_form_helper_path = 'sveedocuments.forms.crispies.insert_helper'
    codemirror_config_name = 'sveetchies-documents-add-insert'
    
    def __init__(self, author=None, *args, **kwargs):
        self.author = author
        
        super(InsertForm, self).__init__(*args, **kwargs)
        super(forms.ModelForm, self).__init__(*args, **kwargs)
        
        extra_settings = {}
        if not self.instance.id:
            # Désactive l'option de sauvegarde rapide
            extra_settings = {'quicksave_url': None}
            
        self.fields['content'].widget = CodeMirrorWidget(attrs={'rows': 30}, config_name=self.codemirror_config_name)
    
    def clean_slug(self):
        slug = self.cleaned_data.get("slug")
        if slug:
            if slug in DOCUMENTS_PAGE_RESERVED_SLUGS:
                raise forms.ValidationError(_('This is a reserved keyword'))
        return slug
    
    def clean_content(self):
        """
        Parse le contenu pour vérifier qu'il ne contient par d'erreurs de syntaxe
        """
        content = self.cleaned_data.get("content")
        if content:
            errors = SourceReporter(content)
            if errors:
                raise forms.ValidationError(map(map_parsing_errors, errors))
        return content
    
    def save(self, *args, **kwargs):
        edited = (self.instance and self.instance.id)

        instance = super(InsertForm, self).save(commit=False, *args, **kwargs)
        instance.author = self.author
        instance.save()
        
        # Send a Django signal for page update
        documents_insert_update_signal.send(sender=self, insert_instance=instance, edited=edited)
        
        return instance
    
    class Meta:
        model = Insert
        exclude = ('created', 'author')
        widgets = {
            # Necessary to change "self.fields['content'].widget" in __init__
            'content': CodeMirrorWidget(attrs={'rows': 30}), 
        }


class InsertEditForm(InsertForm):
    crispy_form_helper_path = 'sveedocuments.forms.crispies.insert_edit_helper'
    codemirror_config_name = 'sveetchies-documents-edit-insert'


class InsertQuickForm(PageQuickForm):
    """
    *Insert* quicksave form
    """
    class Meta:
        model = Insert
        fields = ('content',)



# -*- coding: utf-8 -*-
"""
Command line tool to export

DEPRECATED
"""
import re, StringIO

from optparse import OptionValueError, make_option

from django.conf import settings
from django.core.cache import cache
from django.core.management.base import CommandError, BaseCommand

from sveedocuments.models import Page, Insert

class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option("--get_pages", dest="get_pages", default=None, help="Get pages from given slug. Each slug separated by comma, order does matter. Special slug 'ALL' means all pages in their sitemap order.", metavar="SLUGS"),
        make_option("--follows", dest="followed_pages", default=None, help="Specify each page slug to recursively follow for their children in select. Each slug separated by comma. Special slug 'ALL' means all pages.", metavar="SLUGS"),
        make_option("--excludes", dest="excluded_pages", default=None, help="Specify each page slug to excludes (but not their children) in select. Each slug separated by comma.", metavar="SLUGS"),
        make_option("--print", dest="print_docs", action="store_true", default=False, help="Print out the selected documents."),
        make_option("--resume", dest="resume_docs", action="store_true", default=False, help="Print a title list of selected documents."),
        make_option("--to_file", dest="export_to_file", default=None, help="Export selected documents in a file.", metavar="FILEPATH"),
        make_option("--to_github", dest="export_to_github", default=None, help="Export selected documents in a file but with some changes to be suitable with the Github ReSructuredText parser.", metavar="FILEPATH"),
    )
    help = "[DEPRECATED] Command to export documents from Sveetchies-documents"

    def handle(self, *args, **options):
        if len(args) != 0:
            raise CommandError("Command doesn't accept any arguments")
        
        self.export_to_file = options.get('export_to_file')
        self.export_to_github = options.get('export_to_github')
        self.print_docs = options.get('print_docs')
        self.resume_docs = options.get('resume_docs')
        self.get_pages = options.get('get_pages')
        self.followed_pages = options.get('followed_pages')
        self.excluded_pages = options.get('excluded_pages')
        self.verbosity = int(options.get('verbosity'))
        
        self.compile_hack_regex()
        
        if self.print_docs or self.resume_docs:
            output = StringIO.StringIO()
            self.output_documents(output, resumed=self.resume_docs)
            print output.getvalue()
            output.close()
        
        if self.export_to_file or self.export_to_github:
            output = open(self.export_to_file or self.export_to_github, "w")
            self.output_documents(output)
            output.close()

    def compile_hack_regex(self):
        """
        Compile all regex used for syntax hacks
        
        ``_page_role_regex`` to match : ::
        
            :page:`VALUE`
            
        ``_sourcecode_directive_regex`` to match : ::
        
            ..  sourcecode:: python
                :linenos:
                :hl_lines: 1,2,3
        """
        self._page_role_regex = re.compile(r"(?:\:page\:`)(?P<name>.*?)(?:`)")
        
        simple = r"(?:\s)"
        with_linenos = r"(?:\s[\ ]+\:linenos\:)"
        with_hl_lines = r"(?:\s[\ ]+\:linenos\:\s[\ ]+\:hl_lines\:[\ ][1-9,]+)"
        self._sourcecode_directive_regex = re.compile(r"(?:..[\ ]+sourcecode\:\:[\ ]+)(?P<language>.*?)(?:" + with_hl_lines + r"|" + with_linenos + r"|" + simple + r")")

    def get_documents(self):
        """
        Get all documents matched by given options
        """
        instances = []
        
        if self.get_pages:
            followed = [v for v in (self.followed_pages or '').split(',') if v]
            excluded = [v for v in (self.excluded_pages or '').split(',') if v]
            if self.verbosity:
                print "Followed slugs:", followed
                print "Excluded slugs:", excluded
            
            # Recherche des enfants des pages à exclure
            if excluded:
                tmp = []
                for item in excluded:
                    try:
                        p = Page.objects.get(slug=item).get_descendants(include_self=True).values_list('slug', flat=True)
                    except Page.DoesNotExist:
                        raise CommandError("Excludes: slug '{0}' does not exist".format(item))
                    else:
                        tmp.extend(p)
                    
                excluded = tmp
                if self.verbosity:
                    print "Excluded pages:", excluded
                
            # Slugs des pages à extraire
            if self.get_pages == 'ALL':
                page_slugs = Page.objects.root_nodes().exclude(visible=False, slug__in=excluded).values_list('slug', flat=True)
            else:
                page_slugs = [s for s in self.get_pages.split(',') if s not in excluded]
            
            # Récupère chaque page
            for slug in page_slugs:
                # Si mode suivi intégrale des enfants, ou que le slug de la page est explicitement spécifié à suivre
                if followed and ('ALL' in followed or slug in followed):
                    try:
                        queryset = Page.objects.get(slug=slug).get_descendants(include_self=True).exclude(id__in=[i.id for i in instances]).exclude(visible=False, slug__in=excluded)
                    except Page.DoesNotExist:
                        raise CommandError("Get many: slug '{0}' does not exist".format(slug))
                    else:
                        instances.extend(queryset)
                # Pas de suivi on récupère que la page
                else:
                    try:
                        page = Page.objects.get(slug=slug, visible=True)
                    except Page.DoesNotExist:
                        raise CommandError("Get single: slug '{0}' does not exist".format(slug))
                    else:
                        instances.append(page)
        
        if self.verbosity:
            print "Pages instances:", instances
        if not instances:
            raise CommandError("No document finded")
        return instances

    def output_documents(self, output, resumed=False):
        """
        Export all matched documents in a file
        
        Documents are ordered and their titles are added at top of each them as the 
        highest title
        """
        loaded_docs = self.get_documents()
        from_multiple_docs = len(loaded_docs)>1
        
        for document in loaded_docs:
            content = document.content
            
            if not resumed:
                # Apply hacks on content if needeed
                if self.export_to_github:
                    content = self._page_role_regex.sub(self._page_role_sub_replacement, content)
                    content = self._sourcecode_directive_regex.sub(self._sourcecode_directive_sub_replacement, content)
                # Output the document title + content
                if from_multiple_docs:
                    output.write( "\n" )
                    output.write( self._to_rest_title(document.title.encode('UTF8').strip(), character="=")+"\n\n" )
                output.write( content.encode('UTF8')+"\n" )
            else:
                # Output the document title only
                output.write( "* /%s/ : %s\n" % (document.slug.encode('UTF8'), document.title.encode('UTF8')) )
        
        return output

    def _to_rest_title(self, title, character="="):
        """Transform the document title to a ReST TOC title"""
        width = max([len(line) for line in title.splitlines()])
        ascii_line = character*width
        return ascii_line+"\n"+title+"\n"+ascii_line

    def _page_role_sub_replacement(self, matchobj):
        """Transform page roles to strong emphasis"""
        return u"**{name}**".format(name=matchobj.group(1))

    def _sourcecode_directive_sub_replacement(self, matchobj):
        """Transform sourcecode directives to pre"""
        return u"::"
        